/*
ttf2png - True Type Font to PNG converter
Copyright (c) 2004 Mikkosoft Productions

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <png.h>
#include <ft2build.h>
#include FT_FREETYPE_H

void usage();
int save_png(char *, char *, int, int);

char	verbose=0;
char	alpha=0;

int main(int argc, char **argv)
{
	FT_Library	freetype;
	FT_Face		face;
	int	err;
	int	o;
	int	begin=0;
	int	end=255;
	char	*fn;
	int	size=10;
	int	cpl=16;
	int	w,h;
	int	i;
	char	*data;
	int	ch,cw;
	int	cell=16;
	int	ascent;
	char	*out="font.png";
	const char	*name;
	char	autohinter=0;
	int	count;

	if(argc<2)
	{
		usage();
		return 1;
	}

	while((o=getopt(argc,argv,"r:s:l:c:o:atvh?"))!=-1)
	{
		char	*ptr;
		int	temp;
		switch(o)
		{
		case 'r':
			if(!isdigit(optarg[0])) temp=-1;
			else
			{
				temp=strtol(optarg,&ptr,0);
				if(ptr[0]!=',' || !isdigit(ptr[1])) temp=-1;
			}
			if(temp<0)
				printf("Not a valid range: %s\n",optarg);
			else
			{
				begin=temp;
				end=strtol(ptr+1,NULL,0);
			}
			break;
		case 's':
			size=strtol(optarg,NULL,0);
			break;
		case 'l':
			cpl=strtol(optarg,NULL,0);
			break;
		case 'c':
			cell=strtol(optarg,NULL,0);
			break;
		case 'o':
			out=optarg;
			break;
		case 'a':
			autohinter=1;
			break;
		case 't':
			alpha=1;
			break;
		case 'v':
			++verbose;
			break;
		case 'h':
		case '?':
			usage();
			return 0;
		}
	}

	if(optind>=argc)
	{
		usage();
		return 1;
	}
	fn=argv[optind];

	err=FT_Init_FreeType(&freetype);
	if(err)
	{
		fprintf(stderr,"Couldn't initialize FreeType library\n");
		return 1;
	}

	err=FT_New_Face(freetype,fn,0,&face);
	if(err)
	{
		fprintf(stderr,"Couldn't load font file\n");
		if(err==FT_Err_Unknown_File_Format)
			fprintf(stderr,"Unknown file format\n");
		return 1;
	}

	name=FT_Get_Postscript_Name(face);
	if(verbose)
	{
		printf("Font name: %s\n",name);
		printf("Glyphs: %ld\n",face->num_glyphs);
	}

	err=FT_Set_Pixel_Sizes(face,0,size);
	if(err)
	{
		fprintf(stderr,"Couldn't set size\n");
		return 1;
	}
	ascent=(face->bbox.yMax*face->size->metrics.x_scale)>>16;
	ascent=(ascent+63)/64;
	ch=((face->bbox.yMax-face->bbox.yMin)*face->size->metrics.y_scale)>>16;
	ch=(ch+63)/64;
	cw=((face->bbox.xMax-face->bbox.xMin)*face->size->metrics.x_scale)>>16;
	cw=(cw+63)/64;
	if(verbose)
	{
		printf("Ascent %d\n",ascent);
		printf("Descent %ld\n",(((face->bbox.yMin*face->size->metrics.y_scale)>>16)+63)/64);
		printf("Max height %d\n",ch);
		printf("Max width %d\n",cw);
	}
	if(verbose>=1 && (ch>cell || cw>cell)) fprintf(stderr,"Warning: character size exceeds cell size\n");

	w=cpl*cell;
	h=(end-begin+cpl-1)/cpl*cell;
	data=(char *)malloc(w*h);
	memset(data,255,w*h);

	count=0;
	for(i=begin;i<=end;++i)
	{
		int			glyph=FT_Get_Char_Index(face,i);
		FT_Bitmap	*bmp=&face->glyph->bitmap;
		int			x,y;
		int			cx=(i%cpl)*cell;
		int			cy=(i/cpl)*cell;
		int			flags=0;
		int			dy;
		if(!glyph) continue;
		fflush(stdout);
		if(autohinter) flags|=FT_LOAD_FORCE_AUTOHINT;
		FT_Load_Glyph(face,glyph,flags);
		FT_Render_Glyph(face->glyph,FT_RENDER_MODE_NORMAL);
		if(cell>bmp->width) cx+=(cell-bmp->width)/2;
		dy=ascent-face->glyph->bitmap_top;
		cy+=dy;
		if(verbose>=2)
		{
			printf("  Char %d: glyph %d, size %dx%d\n",i,glyph,bmp->width,bmp->rows);
			if(bmp->width>cell || dy+bmp->rows>cell || dy<0) printf("  Warning: Character %d does not fit in cell\n",i);
		}
		if(bmp->pitch<0)
		{
			fprintf(stderr,"Warning: Character %d not rendered (can't handle reversed bitmaps)\n",i);
			continue;
		}
		for(y=0;y<bmp->rows;++y) for(x=0;x<bmp->width;++x)
		{
			if(cx+x<0 || cx+x>=w || cy+y<0 || cy+y>=h) continue;
			data[cx+x+(cy+y)*w]=255-bmp->buffer[x+y*bmp->pitch];
		}
		++count;
	}

	save_png(out,data,w,h);

	if(verbose) printf("Converted %d glyphs\n",count);

	return 0;
}

void usage()
{
	printf("ttf2png - True Type Font to PNG converter\n"
		"Copyright (c) 2004 Mikkosoft Productions\n"
		"Distributed under the GNU General Public License\n\n"
		"Usage: ttf2png [options] <TTF file>\n\n"
		"Accepted options (default values in [brackets])\n"
		"  -r  Range of characters to convert in the format low,high [0,255]\n"
		"  -s  Font size to use, in pixels [10]\n"
		"  -l  Number of characters to put in one line [16]\n"
		"  -c  Character cell size, in pixels [16]\n"
		"  -o  Output file name [font.png]\n"
		"  -a  Force autohinter\n"
		"  -t  Render font to alpha channel\n"
		"  -v  Increase the level of verbosity\n"
		"  -h  Print this message\n");
}

int save_png(char *fn, char *data, int w, int h)
{
	FILE			*out;
	png_struct	*pngs;
	png_info		*pngi;
	png_byte		*rows[h];
	int			i;
	png_byte		*data2;
	int			color;

	out=fopen(fn,"wb");
	if(!out)
	{
		fprintf(stderr,"Couldn't open output file\n");
		return -1;
	}

	pngs=png_create_write_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
	if(!pngs)
	{
		fprintf(stderr,"Error writing PNG file\n");
		return -1;
	}
	pngi=png_create_info_struct(pngs);
	if(!pngi)
	{
		png_destroy_write_struct(&pngs,NULL);
		fprintf(stderr,"Error writing PNG file\n");
		return -1;
	}

	png_init_io(pngs,out);
	if(alpha)
	{
		data2=(png_byte *)malloc(w*h*2);
		for(i=0;i<w*h;++i)
		{
			data2[i*2]=255;
			data2[i*2+1]=255-data[i];
		}
		for(i=0;i<h;++i)
			rows[i]=(png_byte *)(data2+i*w*2);
		color=PNG_COLOR_TYPE_GRAY_ALPHA;
	}
	else
	{
		for(i=0;i<h;++i)
			rows[i]=(png_byte *)(data+i*w);
		color=PNG_COLOR_TYPE_GRAY;
	}
	png_set_IHDR(pngs,pngi,w,h,8,color,PNG_INTERLACE_NONE,PNG_COMPRESSION_TYPE_DEFAULT,PNG_FILTER_TYPE_DEFAULT);
	png_set_rows(pngs,pngi,rows);
	png_write_png(pngs,pngi,PNG_TRANSFORM_IDENTITY,NULL);
	png_destroy_write_struct(&pngs,&pngi);
	if(alpha) free(data2);

	if(verbose) printf("Saved %dx%d PNG image to %s\n",w,h,fn);

	return 0;
}
